from datetime import datetime, date, timedelta
from enum import Enum
from itertools import groupby
from pathlib import Path
import shutil
from uuid import UUID, uuid4
import sys
from typing import Optional, Set

from ..tasks.task_wrapper import TaskWrapper
from ..utils.dag import query as dag_query, topological_sort
from .settings import get_connections, get_settings
from .errors import Err, Exc, Ok, Result, SaynError
from ..logging import EventTracker
from ..database import Database

from ..core.project import read_project, read_groups, get_tasks_dict
from ..core.settings import read_settings
from ..database.unknown import UnknownDb
from ..database.objects import DbObjectCompiler
from ..logging import ConsoleLogger
from ..utils.python_loader import PythonLoader
from ..utils.task_query import get_query
from ..utils.compiler import Compiler

from ..tasks.task import TaskStatus
from ..tasks.dummy import DummyTask
from ..tasks.sql import SqlTask

from ..tasks.autosql import AutoSqlTask
from ..tasks.copy import CopyTask
from ..tasks.test import TestTask

_creators = {
    "dummy": DummyTask,
    "sql": SqlTask,
    "autosql": AutoSqlTask,
    "copy": CopyTask,
    "test": TestTask,
}


run_id = uuid4()


class Command(Enum):
    UNDEFINED = "undefined"
    COMPILE = "compile"
    RUN = "run"
    TEST = "test"


class RunArguments:
    class Folders:
        python: str = str(Path("python"))
        sql: str = str(Path("sql"))
        compile: str = str(Path("compile"))
        logs: str = str(Path("logs"))
        tests: str = str(Path("sql"))

    folders: Folders
    full_load: bool = False
    start_dt: date = date.today() - timedelta(days=1)
    end_dt: date = date.today() - timedelta(days=1)
    dates_specified: bool = False
    debug: bool = False
    profile: Optional[str] = None
    command: Command = Command.UNDEFINED
    upstream_prod: bool = False
    is_prod: bool = False

    include: Set[str]
    exclude: Set[str]

    def __init__(self):
        self.folders = self.Folders()
        self.include = set()
        self.exclude = set()

    def update(self, **kwargs):
        if "command" in kwargs:
            if isinstance(kwargs["command"], str):
                if kwargs["command"] == "compile":
                    self.command = Command.COMPILE
                elif kwargs["command"] == "run":
                    self.command = Command.RUN
                elif kwargs["command"] == "test":
                    self.command = Command.TEST
            else:
                self.command = kwargs["command"]

        if "debug" in kwargs:
            self.debug = kwargs["debug"]


class App:
    def __init__(self):
        self.project_root = Path(".")

        self.run_id: UUID = run_id
        self.app_start_ts = datetime.now()

        self.run_arguments = RunArguments()

        self.tracker = EventTracker(self.run_id)
        self.tracker.register_logger(ConsoleLogger(True))

        self.project_parameters = dict()
        self.prod_project_parameters = dict()
        self.credentials = dict()
        self.default_db = None

        self.tasks = dict()
        self.dag = dict()
        self.tests = dict()

        self.task_query = list()
        self.tasks_to_run = dict()

        self.connections = dict()

        self.python_loader = PythonLoader()

    def start_app(self):
        self.tracker.report_event(
            context="app",
            event="start_app",
            debug=self.run_arguments.debug,
            full_load=self.run_arguments.full_load,
            start_dt=self.run_arguments.start_dt,
            end_dt=self.run_arguments.end_dt,
            profile=self.run_arguments.profile,
        )
        self.cleanup_compilation()

        # SETUP THE APP: read project config and settings, interpret cli arguments and setup the dag
        self.tracker.start_stage("config")

        # Set python environment
        if Path(self.run_arguments.folders.python).is_dir():
            self.check_abort(
                self.python_loader.register_module(
                    "python_tasks", self.run_arguments.folders.python
                )
            )

        # Read the project configuration
        try:
            project = read_project(self.project_root)
        except SaynError as exc:
            self.finish_app(error=Exc(exc))

        try:
            file_groups = read_groups(self.project_root)
        except SaynError as exc:
            self.finish_app(error=Exc(exc))

        self.set_project(project, file_groups)

        # We need the settings before we can process the tasks
        settings = self.check_abort(read_settings())
        self.check_abort(self.set_settings(settings))

        # Set tasks and dag from it
        tasks_dict = self.check_abort(
            get_tasks_dict(
                self.presets,
                self.file_groups,
                self.autogroups,
                self.run_arguments.folders.sql,
                self.compiler,
                self.python_loader,
            )
        )

        # Set the tasks for the project and call their config method

        if self.run_arguments.command != Command.TEST:
            tasks_dict = {k: v for k, v in tasks_dict.items() if v["type"] != "test"}

        self.check_abort(self.set_tasks(tasks_dict))

        self.tracker.finish_current_stage(
            tasks={k: v.status for k, v in self.tasks.items()},
            test=True if self.run_arguments.command == Command.TEST else False,
        )

        # Setup stage
        self.tracker.start_stage("setup")

        self.check_abort(self.setup_execution())

        self.tracker.finish_current_stage(
            tasks={k: v.status for k, v in self.tasks.items() if v.in_query},
            test=True if self.run_arguments.command == Command.TEST else False,
        )

    def set_project(self, project, file_groups):
        self.prod_project_parameters.update(project.parameters or dict())
        self.project_parameters.update(project.parameters or dict())

        # Temporarily store the raw stringify strings
        self.input_prod_stringify = {
            "database_prefix": project.database_prefix,
            "database_suffix": project.database_suffix,
            "database_override": project.database_override,
            "schema_prefix": project.schema_prefix,
            "schema_suffix": project.schema_suffix,
            "schema_override": project.schema_override,
            "table_prefix": project.table_prefix,
            "table_suffix": project.table_suffix,
            "table_override": project.table_override,
        }
        self.input_stringify = {
            "database_prefix": project.database_prefix,
            "database_suffix": project.database_suffix,
            "database_override": project.database_override,
            "schema_prefix": project.schema_prefix,
            "schema_suffix": project.schema_suffix,
            "schema_override": project.schema_override,
            "table_prefix": project.table_prefix,
            "table_suffix": project.table_suffix,
            "table_override": project.table_override,
        }
        self.credentials = {k: None for k in project.required_credentials}
        self.default_db = project.default_db

        self.presets = project.presets or dict()

        # Validate groups
        collision = set(file_groups.keys()).intersection(set(project.autogroups.keys()))
        if len(collision) > 0:
            if len(collision) == 1:
                error_message = (
                    f'The group "{", ".join(collision)}" is defined both in '
                    '"project.yaml" and as a file in the "tasks" folder: '
                )
            else:
                error_message = (
                    f'Some groups ({", ".join(collision)}) are defined both in '
                    '"project.yaml" and as files in the "tasks" folder: '
                )
            self.finish_app(
                error=Err(
                    "dag",
                    "duplicate_groups",
                    error_message=error_message,
                    groups=list(collision),
                )
            )
        self.autogroups = project.autogroups
        self.file_groups = file_groups

    def set_settings(self, settings):
        settings_dict = get_settings(
            settings["yaml"], settings["env"], self.run_arguments.profile
        )
        if settings_dict.is_err:
            return settings_dict
        else:
            settings_dict = settings_dict.value

        parameters = settings_dict["parameters"] or dict()
        credentials = settings_dict["credentials"] or dict()
        stringify = settings_dict["stringify"] or dict()
        self.from_prod = settings_dict["from_prod"]

        if len(parameters) == 0 and len(stringify) == 0:
            self.run_arguments.is_prod = True

        # Validate the given parameters
        error_items = set(parameters.keys()) - set(self.project_parameters.keys())
        if error_items:
            return Err(
                "app",
                "wrong_parameters",
                parameters=error_items,
            )

        self.project_parameters.update(parameters)

        # With the parameters in place we can build our jinja compiler
        self.compiler = Compiler(
            self.run_arguments, self.project_parameters, self.prod_project_parameters
        )

        # Validate credentials
        error_items = set(credentials.keys()) - set(self.credentials.keys())
        if error_items:
            return Err("app", "wrong_credentials", credentials=error_items)

        error_items = [n for n, v in credentials.items() if "type" not in v]
        if error_items:
            return Err("app", "missing_credential_type", credentials=error_items)

        self.credentials.update(credentials)

        # Create connections
        result = get_connections(
            self.credentials,
        )
        if result.is_err:
            return result
        else:
            self.connections = result.value
        # Object compilation objects
        self.input_stringify.update(stringify)

        self.db_object_compiler = DbObjectCompiler(
            self.connections,
            self.default_db,
            self.input_stringify,
            self.input_prod_stringify,
            self.from_prod,
        )

        # Check the default_run setting and update run arguments
        self.run_arguments.include.update(settings_dict["default_run"]["include"])
        self.run_arguments.exclude.update(settings_dict["default_run"]["exclude"])
        if settings_dict["default_run"]["upstream_prod"] is not None:
            self.run_arguments.upstream_prod = settings_dict["default_run"][
                "upstream_prod"
            ]

        return Ok()

    def get_task_class(self, task_type, config):
        if task_type == "python_module":
            return Ok(config.pop("task_class"))
        elif task_type == "python":
            return self.python_loader.get_class("python_tasks", config.get("class"))
        elif task_type in _creators:
            return Ok(_creators[task_type])
        else:
            return Err(
                "task_type",
                "invalid_task_type_error",
                type=task_type,
                group=config["group"],
            )

    def set_tasks(self, tasks):
        # We first need to do the config of tasks
        failed_tasks = list()
        task_objects = dict()

        if len(tasks) == 0:
            self.finish_app(Err("dag", "empty_dag"))

        for task_name, task in tasks.items():
            task_tracker = self.tracker.get_task_tracker(task_name)
            task_tracker._report_event("start_stage")
            start_ts = datetime.now()

            result = self.get_task_class(task["type"], task)
            if result.is_err:
                task_class = None
                result_error = result
                failed_tasks.append(task_name)
            else:
                task_class = result.value
                result_error = None

            task_objects[task_name] = TaskWrapper(
                task["group"],
                task_name,
                task["type"],
                task.get("on_fail"),
                task.get("parents"),
                task.get("sources"),
                task.get("outputs"),
                task.get("tags"),
                task_tracker,
                task_class,
                self.connections,
                self.default_db,
                self.run_arguments,
                self.compiler,
                self.db_object_compiler,
            )

            if task_class is None:
                task_tracker._report_event(
                    "finish_stage",
                    duration=datetime.now() - start_ts,
                    result=result_error,
                )

            else:
                result = task_objects[task_name].config(
                    task,
                    self.project_parameters,
                    task.get("parameters"),
                )

                if result.is_err:
                    failed_tasks.append(task_name)

                task_tracker._report_event(
                    "finish_stage", duration=datetime.now() - start_ts, result=result
                )

        if len(failed_tasks) > 0:
            # If any tasks fail to do config, we can't ensure the DAG is correct, so we abort
            self.tasks = task_objects
            self.tracker.finish_current_stage(
                tasks={k: v.status for k, v in self.tasks.items()}
            )

            self.finish_app()

        # Now that all tasks are configured, we set the relationships so that we
        # can calculate the dag
        output_to_task = [
            (output, task_name)
            for task_name, task in task_objects.items()
            for output in task.outputs
        ]

        output_to_task = {
            k: [gg[1] for gg in g]
            for k, g in groupby(sorted(output_to_task), key=lambda x: x[0])
        }
        for task_name, task in task_objects.items():
            result = task.set_parents(task_objects, output_to_task)
            if result.is_err:
                return result

        self.dag = {
            task.name: [p.name for p in task.parents] for task in task_objects.values()
        }

        topo_sort = topological_sort(self.dag)
        if topo_sort.is_err:
            return topo_sort
        else:
            topo_sort = topo_sort.value

        self.tasks = {task_name: task_objects[task_name] for task_name in topo_sort}

        return Ok()

    def setup_execution(self):
        # Apply the task query
        tasks_dict = {
            name: {"group": task.group, "tags": list(task.tags)}
            for name, task in self.tasks.items()
        }
        self.task_query = self.check_abort(
            get_query(
                tasks_dict,
                include=self.run_arguments.include,
                exclude=self.run_arguments.exclude,
            )
        )

        result = dag_query(self.dag, self.task_query)
        if result.is_err:
            return result
        else:
            tasks_in_query = result.value

        if self.run_arguments.command == Command.TEST:
            tasks_in_query = [t for t in tasks_in_query if self.tasks[t].has_tests()]

        # Introspection
        #########

        # Create the list of objects to be used in this execution
        exec_outputs = set()
        exec_sources = set()
        exec_connections = set()
        for task_name in tasks_in_query:
            for output in self.tasks[task_name].outputs:
                exec_outputs.add(output)
                # if output.connection not in objects_used:
                #     objects_used[output.connection] = set()
                # objects_used[output.connection].add(output)

            for source in self.tasks[task_name].sources:
                exec_sources.add(source)
                # if source.connection not in objects_used:
                #     objects_used[source.connection] = set()
                # objects_used[source.connection].add(source)

            for connection in self.tasks[task_name].used_connections:
                exec_connections.add(connection)

        # Now that we have done the config for all tasks and we know which
        # connections are required, check that we have them all
        connections_setup = {
            n for n, v in self.connections.items() if not isinstance(v, UnknownDb)
        }
        error_items = exec_connections - connections_setup
        if error_items:
            return Err("app", "missing_credentials", credentials=error_items)

        if self.run_arguments.upstream_prod:
            sources_from_prod = {s for s in exec_sources if s not in exec_outputs}
        else:
            sources_from_prod = set()

        self.db_object_compiler.set_sources_from_prod(sources_from_prod)
        sources_from_prod = {
            s for s in exec_sources if self.db_object_compiler.is_from_prod(s)
        }

        # Get the objects to introspect
        # We create new DbObjects simply because it's simpler to convert to the correct
        # specification depending on whether it's from_prod or not, but these new objects
        # are not meant to be used anywhere else in the code
        to_introspect = {self.db_object_compiler.src_obj(s) for s in exec_sources}
        to_introspect.update({self.db_object_compiler.out_obj(o) for o in exec_outputs})

        # Reshape the list into dictionaries of connection > database > schema > set of objects
        to_introspect = {
            conn: {
                db
                or "": {
                    sch or "": {v.table or "" for v in ggg}
                    for sch, ggg in groupby(gg, lambda x: x.schema)
                }
                for db, gg in groupby(g, lambda x: x.database)
            }
            for conn, g in groupby(
                sorted(
                    to_introspect,
                    key=lambda x: (
                        x.connection_name,
                        x.database or "",
                        x.schema or "",
                        x.table or "",
                    ),
                ),
                key=lambda x: x.connection_name,
            )
        }

        for connection_name in exec_connections:
            db = self.connections[connection_name]
            if isinstance(db, Database):
                try:
                    db._activate_connection()  # This call creates the engine and tests the connection
                except Exception as exc:
                    return Exc(exc, where="create_connection")
                if connection_name in to_introspect:
                    try:
                        db._introspect(to_introspect[connection_name])
                    except Exception as exc:
                        return Err("database", "introspection", exception=exc)

        self.tracker.set_tasks(tasks_in_query)

        for task_order, task_name in enumerate(tasks_in_query):
            task = self.tasks[task_name]
            task.tracker._task_order = task_order + 1
            if task.status != TaskStatus.READY_FOR_SETUP:
                continue
            #  TODO - review from test
            # for task_name, task in self._tasks_dict.items():
            #     if self.run_arguments["command"] == "test":
            #         parents = []
            #     else:
            #         parents = [self.tasks[p] for p in task.get("parents", list())]

            #     task_tracker = self.tracker.get_task_tracker(task_name)

            start_ts = datetime.now()

            task.tracker._report_event("start_stage")

            result = task.setup(task_name in tasks_in_query, sources_from_prod)

            task.tracker._report_event(
                "finish_stage", duration=datetime.now() - start_ts, result=result
            )

        return Ok()

    # Commands

    def check_abort(self, result):
        """Interpret the result of setup opreations returning the value if `result.is_ok`.

        Setup errors from the cli result in execution abort.

        Args:
          result (sayn.errors.Result): The result of a setup operation
        """
        if result is None or not isinstance(result, Result):
            self.finish_app(error=Err("app_setup", "unhandled_error", result=result))
        elif result.is_err:
            self.finish_app(result)
        else:
            return result.value

    def run(self):
        self.execute_dag()

    def compile(self):
        self.execute_dag()

    def test(self):
        self.execute_dag()

    def execute_dag(self):
        # Execution of relevant tasks
        tasks_in_query = {k: v for k, v in self.tasks.items() if v.in_query}
        self.tracker.start_stage(
            self.run_arguments.command.value, tasks=list(tasks_in_query.keys())
        )

        for task in self.tasks.values():
            if not task.in_query:
                continue

            # We force the run/compile so that the skipped status can be calculated,
            # but we only report if the task is in the query
            # if task.in_query:
            task.tracker._report_event("start_stage")
            start_ts = datetime.now()

            if self.run_arguments.command == Command.RUN:
                result = task.run()
            elif self.run_arguments.command == Command.COMPILE:
                result = task.compile()
            elif self.run_arguments.command == Command.TEST:
                result = task.test()
            else:
                self.finish_app(error=Err("cli", "wrong_command"))

            if task.in_query:
                task.tracker._report_event(
                    "finish_stage", duration=datetime.now() - start_ts, result=result
                )

        self.tracker.finish_current_stage(
            tasks={k: v.status for k, v in tasks_in_query.items()},
            test=True if self.run_arguments.command == Command.TEST else False,
        )

        self.finish_app()

    def finish_app(self, error=None):
        duration = datetime.now() - self.app_start_ts
        if error is not None:
            self.tracker.report_event(
                event="finish_app",
                duration=duration,
                error=error,
            )
            sys.exit(-1)
        else:
            self.tracker.report_event(
                event="finish_app",
                duration=duration,
                tasks={k: v.status for k, v in self.tasks.items()},
            )
            if any(
                [
                    t.status in (TaskStatus.SETUP_FAILED, TaskStatus.FAILED)
                    for _, t in self.tasks.items()
                ]
            ):
                sys.exit(-1)
            else:
                sys.exit()

    def cleanup_compilation(self):
        folder = self.run_arguments.folders.compile
        compile_path = Path(folder)
        if compile_path.exists():
            if compile_path.is_dir():
                shutil.rmtree(compile_path.absolute())
            else:
                compile_path.unlink()

        compile_path.mkdir()



import datetime
import decimal

from sqlalchemy import create_engine
from sqlalchemy.sql import sqltypes

from . import Database

db_parameters = ["host", "user", "password", "port", "database"]


class Mysql(Database):
    def feature(self, feature):
        return feature in (
            "CAN REPLACE VIEW",
            "CANNOT SPECIFY DDL IN SELECT",
            "TABLE RENAME CHANGES SCHEMA",
        )

    def create_engine(self, settings):
        # Create engine using the connect_args argument to create_engine
        if "connect_args" not in settings:
            settings["connect_args"] = dict()
        for param in db_parameters:
            if param in settings:
                if param == "port":
                    value = int(settings.pop(param))
                else:
                    value = settings.pop(param)
                settings["connect_args"][param] = value

        return create_engine("mysql+pymysql://", **settings)

    def _list_databases(self):
        """List the accessible databases for this connection."""
        databases = self.read_data("SHOW DATABASES;")
        databases = [db["name"].lower() for db in databases] + [""]
        return databases

    def execute(self, script):
        for s in script.split(";"):
            if len(s.strip()) > 0:
                self.engine.execute(s)

    def _py2sqa(self, from_type):
        python_types = {
            int: sqltypes.BigInteger,
            str: sqltypes.Text,
            float: sqltypes.Float,
            decimal.Decimal: sqltypes.Numeric,
            datetime.datetime: sqltypes.TIMESTAMP,
            bytes: sqltypes.LargeBinary,
            bool: sqltypes.Boolean,
            datetime.date: sqltypes.Date,
            datetime.time: sqltypes.Time,
            datetime.timedelta: sqltypes.Interval,
            list: sqltypes.ARRAY,
            dict: sqltypes.JSON,
        }

        if from_type not in python_types:
            raise ValueError(f'Type not supported "{from_type}"')
        else:
            return python_types[from_type]().compile(dialect=self.engine.dialect)



from pathlib import Path
import importlib
import inspect
import sys

from ..core.errors import Err, Ok


class PythonLoader:
    modules = list()

    def has_module(self, key):
        return key in self.modules

    def register_module(self, key, folder):
        path = Path(folder, "__init__.py")
        if not path.is_file():
            return Err("python_loader", "missing_init_py", path=path)

        loader = importlib.machinery.SourceFileLoader(
            key, str(Path(folder, "__init__.py"))
        )
        spec = importlib.util.spec_from_loader(key, loader)
        m = importlib.util.module_from_spec(spec)

        sys.modules[f"sayn_{key}"] = m

        try:
            loader.exec_module(m)
        except Exception as e:
            return Err(
                "python_loader", "exec_module_exception", module_key=key, exception=e
            )

        self.modules.append(f"sayn_{key}")

        return Ok()

    def get_class(self, module_key, class_path):
        if f"sayn_{module_key}" not in self.modules:
            return Err("python_loader", "module_not_registered", module_key=module_key)
        try:
            module_str = ".".join(class_path.split(".")[:-1])
            class_name = class_path.split(".")[-1]
        except Exception as e:
            return Err("python_loader", "load_class_exception", exception=e)

        if len(module_str) > 0:
            try:
                task_module = importlib.import_module(f"sayn_{module_key}.{module_str}")
            except Exception as e:
                return Err(
                    "python_loader",
                    "load_class_exception",
                    module_key=module_key,
                    exception=e,
                )
        else:
            task_module = importlib.import_module(f"sayn_{module_key}")

        try:
            klass = getattr(task_module, class_name)
        except:
            # module_file_name = (
            #     module_str.replace(".", "/") if len(module_str) > 0 else "__init__"
            # )
            return Err(
                "python_loader",
                "missing_class",
                module_key=module_key,
                module_path=module_str,
                pyclass=class_name,
            )

        return Ok(klass)

    def get_objects(self, module_key, module_str, type_filter=None):
        if f"sayn_{module_key}" not in self.modules:
            return Err("python_loader", "module_not_registered", module_key=module_key)

        try:
            task_module = importlib.import_module(f"sayn_{module_key}.{module_str}")
        except Exception as e:
            return Err("python_loader", "load_class_exception", exception=e)

        if type_filter is None:
            return Ok([func for name, func in inspect.getmembers(task_module)])
        else:
            return Ok(
                [
                    func
                    for _, func in inspect.getmembers(
                        task_module, lambda x: isinstance(x, type_filter)
                    )
                ]
            )



from datetime import datetime, timedelta
from copy import deepcopy
from itertools import groupby


def reverse_dict(dag):
    return {
        i[0]: [ii[1] for ii in i[1]]
        for i in groupby(
            sorted(
                [(parent, node) for node, parents in dag.items() for parent in parents]
            ),
            lambda x: x[0],
        )
    }


def reverse_dict_inclusive(dag):
    rev = {
        i[0]: [ii[1] for ii in i[1]]
        for i in groupby(
            sorted(
                [(parent, node) for node, parents in dag.items() for parent in parents]
            ),
            lambda x: x[0],
        )
    }
    rev.update({n: list() for n in dag.keys() if n not in rev})
    return rev


def merge_dicts(into_dict, from_dict):
    if isinstance(into_dict, list):
        if not isinstance(from_dict, list):
            return deepcopy(from_dict)
        else:
            return deepcopy(into_dict + from_dict)
    elif isinstance(into_dict, dict):
        if not isinstance(from_dict, dict):
            return deepcopy(from_dict)
        else:
            output = dict()
            for k in set(into_dict.keys()).union(from_dict.keys()):
                if k in from_dict and k in into_dict:
                    output[k] = merge_dicts(into_dict[k], from_dict[k])
                else:
                    output[k] = deepcopy(into_dict.get(k) or from_dict.get(k))

            return output
    else:
        return from_dict


def merge_dict_list(dict_list):
    into_dict = dict_list[0]
    for from_dict in dict_list[1:]:
        into_dict = merge_dicts(into_dict, from_dict)

    return into_dict


def map_nested(ob, func):
    if isinstance(ob, dict):
        return {k: map_nested(v, func) for k, v in ob.items()}
    elif isinstance(ob, list):
        return [map_nested(e, func) for e in ob]
    else:
        return func(ob)


def group_list(items):
    return {
        k: [vv[1] for vv in v]
        for k, v in groupby(sorted(items, key=lambda x: x[0]), lambda x: x[0])
    }



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sayn',
 'sayn.core',
 'sayn.database',
 'sayn.logging',
 'sayn.scaffolding',
 'sayn.scaffolding.data.init_project.python',
 'sayn.tasks',
 'sayn.utils']

package_data = \
{'': ['*'],
 'sayn.database': ['templates/*'],
 'sayn.scaffolding': ['data/init_project/*', 'data/init_project/sql/*'],
 'sayn.tasks': ['tests/*']}

install_requires = \
['Click==8.0.4',
 'Jinja2>=3.0.3',
 'SQLAlchemy>=1.4.25,<2.0.0',
 'colorama>=0.4.4',
 'halo==0.0.31',
 'orjson>=3.7.7',
 'pydantic>=1.9.0,<1.10.0',
 'ruamel.yaml.clib==0.2.7',
 'ruamel.yaml==0.17.21']

extras_require = \
{':extra == "redshift"': ['redshift-connector>=2.0.910,<3.0.0',
                          'sqlalchemy-redshift>=0.8.12,<0.9.0'],
 'all': ['psycopg2>=2.9.3',
         'PyMySQL>=1.0.2',
         'snowflake-sqlalchemy>=1.4.4',
         'graphviz>=0.19.1',
         'graphviz>=0.19.1'],
 'all:python_full_version >= "3.8.1" and python_version < "3.12"': ['sqlalchemy-bigquery>=1.6.0'],
 'bigquery': ['google-cloud-bigquery-storage>=2.0.0,<3.0.0dev'],
 'bigquery:python_full_version >= "3.8.1" and python_version < "3.12"': ['sqlalchemy-bigquery>=1.6.0'],
 'graphviz': ['graphviz>=0.19.1'],
 'mysql': ['PyMySQL>=1.0.2'],
 'postgresql': ['psycopg2>=2.9.3'],
 'snowflake': ['snowflake-sqlalchemy>=1.4.4']}

entry_points = \
{'console_scripts': ['sayn = sayn.cli:cli']}

setup_kwargs = {
    'name': 'sayn',
    'version': '0.6.8',
    'description': 'Data-modelling and processing framework for automating Python and SQL tasks',
    'long_description': '<img\n  src="https://173-static-files.s3.eu-west-2.amazonaws.com/sayn_docs/logos/sayn_logo.png"\n  alt="SAYN logo"\n  style="width: 50%; height: 50%;"\n/>\n\n#\n\nSAYN is a modern data processing and modelling framework. Users define tasks (incl. Python, automated SQL transformations and more) and their relationships, SAYN takes care of the rest. It is designed for simplicity, flexibility and centralisation in order to bring significant efficiency gains to the data engineering workflow.\n\n## Use Cases\n\nSAYN can be used for multiple purposes across the data engineering and analytics workflows:\n\n* Data extraction: complement tools such as Fivetran or Stitch with customised extraction processes.\n* Data modelling: transform raw data in your data warehouse (e.g. aggregate activity or sessions, calculate marketing campaign ROI, etc.).\n* Data science: integrate and execute data science models.\n\n## Key Features\n\nSAYN has the following key features:\n\n* YAML based DAG (Direct Acyclic Graph) creation. This means all analysts, including non Python proficient ones, can easily add tasks to ETL processes with SAYN.\n* [Automated SQL transformations](https://173tech.github.io/sayn/tasks/autosql/): write your SELECT statement. SAYN turns it into a table/view and manages everything for you.\n* [Jinja parameters](https://173tech.github.io/sayn/parameters/): switch easily between development and product environment and other tricks with Jinja templating.\n* [Python tasks](https://173tech.github.io/sayn/tasks/python/): use Python scripts to complement your extraction and loading layer and build data science models.\n* Multiple [databases](https://173tech.github.io/sayn/databases/overview/) supported.\n* and much more... See the [Documentation](https://173tech.github.io/sayn/).\n\n## Design Principles\n\nSAYN aims to empower data engineers and analysts through its  three core design principles:\n\n* **Simplicity**: data processes should be easy to create, scale and maintain. So your team can focus on data transformation instead of writing processes. SAYN orchestrates all your tasks systematically and provides a lot of automation features.\n* **Flexibility**: the power of data is unlimited and so should your tooling. SAYN supports both SQL and Python so your analysts can choose the most optimal solution for each process.\n* **Centralisation**: all analytics code should live in one place, making your life easier and allowing dependencies throughout the whole analytics process.\n\n## Quick Start\n\nSAYN supports Python 3.7 to 3.10.\n\n```bash\n$ pip install sayn\n$ sayn init test_sayn\n$ cd test_sayn\n$ sayn run\n```\n\nThis is it! You completed your first SAYN run on the example project. Continue with the [Tutorial: Part 1](https://173tech.github.io/sayn/tutorials/tutorial_part1/) which will give you a good overview of SAYN\'s true power!\n\n## Release Updates\n\nIf you want to receive update emails about SAYN releases, you can sign up [here](http://eepurl.com/hnfJIr).\n\n## Support\n\nIf you need any help with SAYN, or simply want to know more, please contact the team at <sayn@173tech.com>.\n\n## License\n\nSAYN is open source under the [Apache 2.0](https://www.apache.org/licenses/LICENSE-2.0) license.\n\n---\n\nMade with :heart: by [173tech](https://www.173tech.com).\n',
    'author': 'Robin Watteaux',
    'author_email': 'robin@173tech.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://173tech.github.io/sayn',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8.1,<=4.0',
}


setup(**setup_kwargs)



from datetime import datetime
from typing import Any, Dict, Optional, List, Union
import re

from pydantic import BaseModel, Field, validator, Extra
from sqlalchemy import or_, select, column

from ..core.errors import Err, Exc, Ok
from ..database import Database
from .sql import SqlTask

# from .test import Columns


class Source(BaseModel):
    supports_schemas: bool
    db_type: str

    db_schema: Optional[str] = Field(None, alias="schema")
    table: str
    db: str
    db_name: Optional[str]

    class Config:
        extra = Extra.forbid

    @validator("db_schema")
    def can_use_schema(cls, v, values):
        if v is not None and not values["supports_schemas"]:
            raise ValueError(
                f'schema not supported for database of type {values["db_type"]}'
            )

        return v


class Destination(BaseModel):
    supports_schemas: bool
    db_type: str

    tmp_schema: Optional[str]
    db_schema: Optional[str] = Field(None, alias="schema")
    table: str
    db: Optional[str]
    db_name: Optional[str]

    class Config:
        extra = Extra.forbid

    @validator("tmp_schema")
    def can_use_tmp_schema(cls, v, values):
        if v is not None and not values["supports_schemas"]:
            raise ValueError(
                f'tmp_schema not supported for database of type {values["db_type"]}'
            )

        return v

    @validator("db_schema")
    def can_use_schema(cls, v, values):
        if v is not None and not values["supports_schemas"]:
            raise ValueError(
                f'schema not supported for database of type {values["db_type"]}'
            )

        return v


class Config(BaseModel):
    source: Source
    destination: Destination
    # ddl: Optional[Dict[str, Any]]
    delete_key: Optional[str]
    append: bool = False
    incremental_key: Optional[str]
    max_merge_rows: Optional[int]
    max_batch_rows: Optional[int]
    columns: Optional[List[Union[str, Dict[str, Any]]]] = list()
    table_properties: Optional[List[Dict[str, Any]]] = list()
    post_hook: Optional[List[Dict[str, Any]]] = list()

    class Config:
        extra = Extra.forbid

    @validator("incremental_key", always=True)
    def incremental_validation(cls, v, values):
        if v is None:  # Full load
            if values.get("delete_key") is not None:
                raise ValueError(
                    'Incremental copy requires both "incremental_key" and "delete_key" or "incremental_key" and "append: true"'
                )
        else:
            if values.get("delete_key") is not None and values.get("append"):
                raise ValueError(
                    '"Append" incremental copy is incompatible with "delete_key"'
                )
            elif values.get("delete_key") is None and not values.get("append"):
                raise ValueError(
                    '"Append" incremental copy requires "delete_key" or  "append: True"'
                )

        return v

    @validator("max_merge_rows")
    def merge_batch_size_val(cls, v, values):
        if values.get("incremental_key") is None:
            raise ValueError("max_merge_rows is only applicable to incremental copy")

        return v


class CopyTask(SqlTask):
    def config(self, **config):  # noqa: C901
        if "task_name" in self._config_input:
            del self._config_input["task_name"]

        conn_names_list = [
            n for n, c in self.connections.items() if isinstance(c, Database)
        ]

        # check the source db exists in settings
        if (
            isinstance(config.get("source"), dict)
            and config["source"].get("db") is not None
        ):
            if config["source"]["db"] not in conn_names_list:
                return Err(
                    "task_definition",
                    "source_db_not_in_settings",
                    db=config["source"]["db"],
                )

        # set the target db for execution
        # this check needs to happen here so we can pass db_features and db_type to the validator
        if (
            isinstance(config.get("destination"), dict)
            and config["destination"].get("db") is not None
        ):
            if config["destination"]["db"] not in conn_names_list:
                return Err(
                    "task_definition",
                    "destination_db_not_in_settings",
                    db=config["destination"]["db"],
                )
            self._target_db = config["destination"]["db"]
        else:
            self._target_db = self._default_db

        if isinstance(config.get("source"), dict):
            config["source"].update(
                {
                    "supports_schemas": not self.connections[
                        config["source"]["db"]
                    ].feature("NO SCHEMA SUPPORT"),
                    "db_type": self.connections[config["source"]["db"]].db_type,
                }
            )

        if isinstance(config.get("destination"), dict):
            config["destination"].update(
                {
                    "supports_schemas": not self.target_db.feature("NO SCHEMA SUPPORT"),
                    "db_type": self.target_db.db_type,
                }
            )

        try:
            self.task_config = Config(**config)
        except Exception as e:
            return Exc(e)

        # Setup sources
        self.source_db = self.connections[self.task_config.source.db]
        if (self.task_config.source.db_name is None) and (
            self.task_config.source.db_schema is None
        ):
            self.source_name = None
            self.source_schema = None
            self.source_table = self.src(
                self.task_config.source.table, connection=self.source_db
            )
        elif (self.task_config.source.db_name is None) and (
            self.task_config.source.db_schema is not None
        ):
            self.source_name = None
            obj = self.src(
                f"{self.task_config.source.db_schema}.{self.task_config.source.table}",
                connection=self.source_db,
            )
            self.source_schema = obj.split(".")[0]
            self.source_table = obj.split(".")[1]
        else:
            obj = self.src(
                f"{self.task_config.source.db_name}.{self.task_config.source.db_schema}.{self.task_config.source.table}",
                connection=self.source_db,
            )
            self.source_name = obj.split(".")[0]
            self.source_schema = obj.split(".")[1]
            self.source_table = obj.split(".")[2]

        # Setup outputs
        self.config_tmp_schema = (
            self.task_config.destination.tmp_schema
            or self.task_config.destination.db_schema
        )
        config_db = self.task_config.destination.db_name
        config_schema = self.task_config.destination.db_schema
        config_table = self.task_config.destination.table
        config_tmp_table = f"sayn_tmp_{config_table}"

        if (config_db is None) and (config_schema is None):
            self.database = None
            self.schema = None
            self.table = self.out(config_table, connection=self.target_db)
        elif (config_db is None) and (config_schema is not None):
            obj = self.out(f"{config_schema}.{config_table}", connection=self.target_db)
            self.database = None
            self.schema = obj.split(".")[0]
            self.table = obj.split(".")[1]
        else:
            obj = self.out(
                f"{config_db}.{config_schema}.{config_table}", connection=self.target_db
            )
            self.database = obj.split(".")[0]
            self.schema = obj.split(".")[1]
            self.table = obj.split(".")[2]

        if self.config_tmp_schema is None:
            self.tmp_schema = None
            self.tmp_table = self.out(config_tmp_table, connection=self.target_db)
        else:
            obj = self.out(
                f"{config_schema}.{config_tmp_table}",
                connection=self.target_db,
            )
            self.tmp_schema = obj.split(".")[0]
            self.tmp_table = obj.split(".")[1]

        self.delete_key = self.task_config.delete_key
        self.src_incremental_key = self.task_config.incremental_key
        self.dst_incremental_key = self.task_config.incremental_key
        self.max_merge_rows = self.task_config.max_merge_rows
        self.max_batch_rows = self.task_config.max_batch_rows

        if self.task_config.append:
            self.mode = "append"
        elif self.dst_incremental_key is None:
            self.mode = "full"
        else:
            self.mode = "inc"

        self.is_full_load = self.run_arguments["full_load"] or self.mode == "full"

        result = self.target_db._validate_ddl(
            self.task_config.columns,
            self.task_config.table_properties,
            self.task_config.post_hook,
        )
        if result.is_ok:
            self.columns = result.value

            # Check if the incremental_key in the destination needs renaming
            if (
                self.dst_incremental_key is not None
                and len(self.columns["columns"]) > 0
            ):
                columns_dict = {
                    c["name"]: c["dst_name"] or c["name"]
                    for c in self.columns["columns"]
                }
                self.dst_incremental_key = columns_dict[self.src_incremental_key]
        else:
            return result

        if self.run_arguments["command"] == "test" and len(self.columns["columns"]) > 0:
            result = self.target_db._construct_tests(
                self.columns["columns"], self.table, self.schema
            )
            if result.is_err:
                return result
            else:
                self.test_query = result.value[0]
                self.test_breakdown = result.value[1]

            if self.test_query is not None:
                self._has_tests = True

        return Ok()

    def setup(self):
        if self.needs_recompile:
            if (self.task_config.source.db_name is None) and (
                self.task_config.source.db_schema is None
            ):
                self.source_name = None
                self.source_schema = None
                self.source_table = self.src(
                    self.task_config.source.table, connection=self.source_db
                )
            elif (self.task_config.source.db_name is None) and (
                self.task_config.source.db_schema is not None
            ):
                self.source_name = None
                obj = self.src(
                    f"{self.task_config.source.db_schema}.{self.task_config.source.table}",
                    connection=self.source_db,
                )
                self.source_schema = obj.split(".")[0]
                self.source_table = obj.split(".")[1]
            else:
                obj = self.src(
                    f"{self.task_config.source.db_name}.{self.task_config.source.db_schema}.{self.task_config.source.table}",
                    connection=self.source_db,
                )
                self.source_name = obj.split(".")[0]
                self.source_schema = obj.split(".")[1]
                self.source_table = obj.split(".")[2]

            if self._has_tests:
                schema = self.task_config.destination.db_schema
                table = self.task_config.destination.table
                obj = self.src(f"{schema}.{table}", self.target_db)
                test_schema = obj.split(".")[0]
                test_table = obj.split(".")[1]
                result = self.target_db._construct_tests(
                    self.ddl["columns"], test_table, test_schema
                )
                if result.is_err:
                    return result
                else:
                    self.test_query = result.value[0]
                    self.test_breakdown = result.value[1]

        return Ok()

    def compile(self):
        result = self.get_columns()
        if result.is_err:
            return result

        return self.execute(False, self.run_arguments["debug"], self.is_full_load)

    def run(self):
        result = self.get_columns()
        if result.is_err:
            return result

        if self.max_merge_rows is not None:
            result = self.execute(
                True,
                self.run_arguments["debug"],
                self.is_full_load,
                self.max_merge_rows,
            )
            if result.is_err:
                return result
            for _ in range(100):
                if result.is_err or result.value < self.max_merge_rows:
                    break
                result = self.execute(True, False, False, self.max_merge_rows)
        else:
            result = self.execute(True, self.run_arguments["debug"], self.is_full_load)

        return result

    def test(self):
        step_queries = {
            "Write Test Query": self.test_query,
            "Execute Test Query": self.test_query,
        }
        breakdown = self.get_test_breakdown(self.test_breakdown)

        if self.test_query is None:
            self.info("Nothing to be done")
            return self.success()
        else:
            self.set_run_steps(list(step_queries.keys()))

            for step, query in step_queries.items():
                with self.step(step):
                    if "Write" in step:
                        self.write_compilation_output(query, "test")
                    if "Execute" in step:
                        try:
                            result = self.default_db.read_data(query)
                        except Exception as e:
                            return Exc(e)

            if len(result) == 0:
                return self.test_sucessful(breakdown)
            else:
                errout, failed = self.test_failure(
                    breakdown, result, self.run_arguments["debug"]
                )
                problematic_values_query = self.default_db.test_problematic_values(
                    failed, self.table, self.schema
                )

                for query in problematic_values_query.split(";"):
                    if query.strip():
                        header = re.search(r"--.*?--", query).group(0)
                        self.info("")
                        self.info(header)
                        self.info(
                            "===================================================================="
                        )
                        self.info(
                            re.sub(r"--.*?--", "", query).replace("\n", " ").strip()
                            + ";"
                        )
                        self.info(
                            "===================================================================="
                        )
                        self.info("")

                self.write_compilation_output(
                    problematic_values_query, "test_problematic_values"
                )
                return errout

    def execute(self, execute, debug, is_full_load, limit=None):
        # Introspect target
        self.target_table_exists = self.target_db._table_exists(self.table, self.schema)

        steps = ["Prepare Load", "Load Data"]
        if self.target_table_exists:
            load_db = self.database
            load_table = self.tmp_table
            load_schema = self.tmp_schema
            if is_full_load or self.mode == "full":
                steps.append("Move Table")
            else:
                steps.append("Merge Tables")
        else:
            load_db = self.database
            load_table = self.table
            load_schema = self.schema

        self.set_run_steps(steps)

        with self.step("Prepare Load"):
            result = self.get_read_query(execute, debug, is_full_load, limit)
            if result.is_err:
                return result
            else:
                get_data_query = result.value

            create_ddl = {k: v for k, v in self.columns.items() if k != "columns"}

            if self.mode == "append":
                create_ddl["columns"] = [c for c in self.columns["columns"]] + [
                    {"name": "_sayn_load_ts", "type": "TIMESTAMP"}
                ]
            else:
                create_ddl["columns"] = [c for c in self.columns["columns"]]

            query = self.target_db.create_table(
                load_table, schema=load_schema, db=load_db, replace=True, **create_ddl
            )
            if debug:
                self.write_compilation_output(query, "create_table")
            if execute:
                try:
                    self.target_db.execute(query)
                except Exception as e:
                    return Exc(e)

        with self.step("Load Data"):
            n_records = 0
            if execute:
                data_iter = self.source_db._read_data_stream(get_data_query)

                def read_iter(iter):
                    if self.mode == "append":

                        load_time = datetime.utcnow()
                        for record in iter:
                            yield dict(record, _sayn_load_ts=load_time)

                    else:
                        for record in iter:
                            yield record

                n_records = self.target_db.load_data(
                    load_table,
                    read_iter(data_iter),
                    db=load_db,
                    schema=load_schema,
                    batch_size=self.max_batch_rows,
                )
        # Final step
        final_step = steps[-1]
        if final_step == "Move Table":
            query = self.target_db.move_table(
                load_table,
                self.table,
                src_schema=load_schema,
                dst_schema=self.schema,
                src_db=self.source_name,
                dst_db=self.database,
                **self.columns,
            )
        elif final_step == "Merge Tables":
            query = self.target_db.merge_tables(
                load_table,
                self.table,
                self.delete_key,
                src_schema=load_schema,
                dst_schema=self.schema,
                src_db=self.source_name,
                dst_db=self.database,
                **self.columns,
            )
        else:
            query = None

        if query is not None:
            with self.step(final_step):
                if debug:
                    self.write_compilation_output(
                        query, final_step.replace(" ", "_").lower()
                    )
                if execute:
                    try:
                        self.target_db.execute(query)
                    except Exception as e:
                        return Exc(e)

        return Ok(n_records)

    def get_columns(self):  # noqa: C901
        # We get the source table definition
        source_table_def = self.source_db._get_table(
            self.source_table,
            self.source_schema,
        )
        if source_table_def is None:
            return Err(
                "database_error",
                "source_db_missing_source_table",
                schema=self.source_schema,
                table=self.source_table,
                db=self.source_db.name,
            )
        self.source_table_def = source_table_def

        if len(self.columns["columns"]) == 0:
            dst_table_def = None
            if not self.is_full_load:
                dst_table_def = self.target_db._get_table(self.table, self.schema)

            if dst_table_def is not None:
                # In incremental loads we use the destination table to determine the columns
                self.columns["columns"] = [
                    {
                        "name": c.name,
                        "type": c.type.compile(dialect=self.target_db.engine.dialect),
                    }
                    for c in dst_table_def.columns
                    if not c.name.startswith("_sayn")
                ]

                # Ensure these columns are in the source
                missing_columns = set(
                    [
                        c.name
                        for c in dst_table_def.columns
                        if not c.name.startswith("_sayn")
                    ]
                ) - set([c.name for c in self.source_table_def.columns])

                if len(missing_columns) > 0:
                    return Err(
                        "database_error",
                        "source_table_missing_columns",
                        db=self.source_db.name,
                        table=self.source_table,
                        schema=self.source_schema,
                        columns=missing_columns,
                    )

            else:
                # In any other case, we use the source
                for c in self.source_table_def.columns:
                    try:
                        col_type = self.target_db._py2sqa(c.type.python_type)
                    except:
                        col_type = c.type.compile(self.target_db.engine.dialect)
                    self.columns["columns"].append({"name": c.name, "type": col_type})
        else:
            # Fill up column types from the source table
            for col in self.columns["columns"]:
                if col.get("name") not in self.source_table_def.columns:
                    return Err(
                        "database_error",
                        "source_table_missing_columns",
                        db=self.source_db.name,
                        table=self.source_table,
                        schema=self.source_schema,
                        column=col.get("name"),
                    )

                if col.get("type") is None:
                    try:
                        col["type"] = self.source_table_def.columns[
                            col["name"]
                        ].type.compile(self.target_db.engine.dialect)
                    except:
                        col["type"] = self.target_db._py2sqa(
                            self.source_table_def.columns[col["name"]].type.python_type
                        )

        for col in self.columns["columns"]:
            col["src_name"] = col["name"]
            if col.get("dst_name") is not None:
                col["name"] = col["dst_name"]

        return Ok()

    def get_read_query(self, execute, debug, is_full_load, limit=None):
        # Get the incremental value
        last_incremental_value_query = (
            f"SELECT MAX({self.dst_incremental_key}) AS value\n"
            f"FROM {'' if self.schema is None else self.schema +'.'}{self.table}\n"
            f"WHERE {self.dst_incremental_key} IS NOT NULL"
        )
        if debug:
            self.write_compilation_output(
                last_incremental_value_query, "last_incremental_value"
            )

        get_data_query = select(
            columns=[
                column(c["src_name"]).label(c["name"])
                if c["src_name"] != c["name"]
                else column(c["src_name"])
                for c in self.columns["columns"]
            ],
            from_obj=self.source_table_def,
        )
        last_incremental_value = None

        if (
            not is_full_load
            and self.target_table_exists
            and self.dst_incremental_key is not None
        ):
            if execute:
                res = self.target_db.read_data(last_incremental_value_query)
                if len(res) == 1:
                    last_incremental_value = res[0]["value"]
            else:
                last_incremental_value = "LAST_INCREMENTAL_VALUE"

        # Select stream
        if last_incremental_value is not None:
            get_data_query = get_data_query.where(
                or_(
                    self.source_table_def.c[self.src_incremental_key].is_(None),
                    self.source_table_def.c[self.src_incremental_key]
                    >= last_incremental_value,
                )
            )

        if self.src_incremental_key is not None:
            get_data_query = get_data_query.order_by(self.src_incremental_key)

        if limit is not None:
            get_data_query = get_data_query.limit(limit)

        if debug:
            try:
                q = get_data_query.compile(compile_kwargs={"literal_binds": True})
            except:
                # compilation can fail when using values like dates
                q = str(get_data_query)
            self.write_compilation_output(q, "get_data")

        return Ok(get_data_query)



import re

from ..core.errors import Err, Ok

#####################################
# Task query interpretation functions
#####################################

RE_TASK_QUERY = re.compile(
    (
        r"^("
        r"(?!(group:|tag:))(?P<upstream>\+?)(?P<task>[a-zA-Z0-9][-_a-zA-Z0-9]+)(?P<downstream>\+?)|"
        r"group:(?P<group>[a-zA-Z0-9][-_a-zA-Z0-9]+)|"
        r"tag:(?P<tag>[a-zA-Z0-9][-_a-zA-Z0-9]+)"
        r")$"
    )
)


def _get_query_component(tasks, query):
    tasks = {
        k: {"group": v["group"], "tags": v.get("tags", list())}
        for k, v in tasks.items()
    }
    match = RE_TASK_QUERY.match(query)
    if match is None:
        return Err(
            "task_query",
            "incorrect_syntax",
            query=query,
        )
    else:
        match_components = match.groupdict()

        if match_components.get("tag") is not None:
            tag = match_components["tag"]
            relevant_tasks = {
                k: v for k, v in tasks.items() if tag in v.get("tags", list())
            }
            if len(relevant_tasks) == 0:
                return Err(
                    "task_query",
                    "undefined_tag",
                    tag=tag,
                )
            return Ok(
                [
                    {"task": task, "upstream": False, "downstream": False}
                    for task in relevant_tasks.keys()
                ]
            )

        if match_components.get("group") is not None:
            group = match_components["group"]
            relevant_tasks = {k: v for k, v in tasks.items() if group == v.get("group")}
            if len(relevant_tasks) == 0:
                return Err("task_query", "undefined_group", group=group)
            return Ok(
                [
                    {"task": task, "upstream": False, "downstream": False}
                    for task in relevant_tasks.keys()
                ]
            )

        if match_components.get("task") is not None:
            task = match_components["task"]
            if task not in tasks:
                return Err(
                    "task_query",
                    "undefined_task",
                    task=task,
                )
            return Ok(
                [
                    {
                        "task": task,
                        "upstream": match_components.get("upstream", "") == "+",
                        "downstream": match_components.get("downstream", "") == "+",
                    }
                ]
            )

        return Err("task_query", "wrong_query")


def get_query(tasks, include=None, exclude=None):
    if include is None:
        include = set()

    if exclude is None:
        exclude = set()

    overlap = set(include).intersection(set(exclude))
    if len(overlap) > 0:
        overlap = ", ".join(overlap)
        return Err(
            "task_query",
            "query_overlap",
            overlap=overlap,
        )

    output = list()
    for operation, components in (("include", include), ("exclude", exclude)):
        for q in components:
            result = _get_query_component(tasks, q)
            if result.is_err:
                return result
            else:
                components = result.value
            for comp in components:
                output.append(dict(comp, operation=operation))

    # simplify the queries by unifying upstream/downstream
    include = dict()
    exclude = dict()
    for operand in output:
        operation_dict = include if operand["operation"] == "include" else exclude
        task = operand["task"]
        if task in operation_dict:
            for flag in ("upstream", "downstream"):
                operation_dict[task][flag] = operation_dict[task][flag] or operand[flag]
        else:
            operation_dict[task] = {
                "upstream": operand["upstream"],
                "downstream": operand["downstream"],
            }

    return Ok(
        [
            dict(flags, task=task, operation=operation)
            for operation, operands in (("include", include), ("exclude", exclude))
            for task, flags in operands.items()
        ]
    )



from abc import ABC, abstractmethod
from copy import deepcopy
from pathlib import Path
from typing import Union

from jinja2 import Environment, FileSystemLoader, StrictUndefined, Template

from ..core.errors import SaynCompileError, SaynMissingFileError


class TaskJinjaEnv:
    """This class is the `task` object when compiling task configurations"""

    def __init__(self, name=None, group=None):
        if group is not None:
            self.group = group
        if name is not None:
            self.name = name


class BaseCompiler(ABC):
    @abstractmethod
    def compile(self, obj: Union[Template, Path, str], **kwargs) -> str:
        pass

    @abstractmethod
    def compile_prod(self, obj: Union[Path, str], **kwargs) -> str:
        pass


class Compiler(BaseCompiler):
    def __init__(self, run_arguments, parameters, prod_parameters):
        env_arguments = {
            "full_load": run_arguments.full_load,
            "start_dt": f"'{run_arguments.start_dt.strftime('%Y-%m-%d')}'",
            "end_dt": f"'{run_arguments.end_dt.strftime('%Y-%m-%d')}'",
        }

        self.env = self._create_environment()
        self.env.globals.update(**env_arguments)
        self.env.globals.update(**parameters)

        self.prod_env = self._create_environment()
        self.prod_env.globals.update(**env_arguments)
        self.prod_env.globals.update(**prod_parameters)

    def _create_environment(self):
        return Environment(
            loader=FileSystemLoader(Path(".")),
            undefined=StrictUndefined,
            keep_trailing_newline=True,
            cache_size=0,
        )

    def _get_template(
        self, obj: Union[Template, Path, str], use_prod: bool = False
    ) -> Template:
        if use_prod:
            env = self.prod_env
        else:
            env = self.env

        if isinstance(obj, Template):
            return obj
        elif isinstance(obj, Path):
            if not obj.is_file():
                raise SaynMissingFileError(str(obj))
            else:
                return env.from_string(obj.read_text(encoding="utf-8"))

        elif isinstance(obj, str):
            return env.from_string(obj)

        else:
            raise SaynCompileError(f'Cannot compile object of type "{type(obj)}"')

    def _compile_template(self, template: Template, **kwargs) -> str:
        out = template.render(**kwargs)
        if out is None:
            return ""
        else:
            return out

    def add_global(self, name, obj):
        self.env.globals[name] = obj
        self.prod_env.globals[name] = obj

    def update_globals(self, **params):
        self.env.globals.update(**params)
        self.prod_env.globals.update(**params)

    def compile(self, obj: Union[Template, Path, str], **kwargs) -> str:
        template = self._get_template(obj, False)
        return self._compile_template(template, **kwargs)

    def compile_prod(self, obj: Union[Path, str], **kwargs) -> str:
        template = self._get_template(obj, False)
        return self._compile_template(template, **kwargs)

    def prepare(self, obj):
        return Prepared(self, self._get_template(obj))

    # Factories
    def get_task_compiler(self, group, name) -> BaseCompiler:
        return TaskCompiler(
            self.env, self.prod_env, TaskJinjaEnv(group=group, name=name)
        )

    # def get_db_object_compiler(self) -> BaseCompiler:
    #     pass


class Prepared:
    """A prepared statement ready to be compiled"""

    def __init__(self, compiler, template):
        self.compiler = compiler
        self.template = template

    def compile(self, **kwargs):
        return self.compiler.compile(self.template, **kwargs)

    def compile_prod(self, **kwargs):
        return self.compiler.compile_prod(self.template, **kwargs)


class TaskCompiler(Compiler):
    def __init__(self, base_env, base_prod_env, task) -> None:
        self.env = self._create_environment()
        self.env.globals.update(**deepcopy(base_env.globals))
        self.env.globals["task"] = task

        self.prod_env = self._create_environment()
        self.prod_env.globals.update(**deepcopy(base_prod_env.globals))
        self.prod_env.globals["task"] = task



from pathlib import Path

from pydantic import BaseModel, FilePath, validator, Extra
from typing import List, Optional, Union
from enum import Enum
from colorama import Fore, Style

from ..core.errors import Ok, Err, Exc
from ..database import Database
from .task import Task


class Tests(BaseModel):
    name: Optional[str]
    values: Optional[List[str]]

    class Config:
        extra = Extra.forbid


class Columns(BaseModel):
    name: str
    description: Optional[str]
    tests: List[Union[str, Tests]]

    class Config:
        extra = Extra.forbid


class Config(BaseModel):
    test_folder: Path
    file_name: FilePath

    class Config:
        extra = Extra.forbid

    @validator("file_name", pre=True)
    def file_name_plus_folder(cls, v, values):
        return Path(values["test_folder"], v)


class OnFailValue(str, Enum):
    skip = "skip"
    no_skip = "no_skip"


class CompileConfig(BaseModel):
    tags: Optional[List[str]]
    # sources: Optional[List[str]]
    # outputs: Optional[List[str]]
    # parents: Optional[List[str]]
    # on_fail: Optional[OnFailValue]

    class Config:
        extra = Extra.forbid


class TestTask(Task):
    def config(self, **config):
        self._has_tests = True

        conn_names_list = [
            n for n, c in self.connections.items() if isinstance(c, Database)
        ]

        # set the target db for execution
        if config.get("db") is not None:
            if config["db"] not in conn_names_list:
                return Err("task_definition", "db_not_in_settings", db=config["db"])
            self._target_db = config["db"]
        else:
            self._target_db = self._default_db

        try:
            self.task_config = Config(
                test_folder=self.run_arguments["folders"]["tests"], **config
            )
        except Exception as e:
            return Exc(e)

        self.compiler.update_globals(
            src=lambda x: self.src(x, connection=self._target_db),
            config=self.config_macro,
        )

        self.test_query = self.compiler.compile(self.task_config.file_name)
        self.test_query += " LIMIT 5\n"

        return Ok()

    def config_macro(self, **config):
        if self.allow_config:
            task_config_override = CompileConfig(**config)

            # Sent to the wrapper
            # if task_config_override.on_fail is not None:
            #     self._config_input["on_fail"] = task_config_override.on_fail

            if task_config_override.tags is not None:
                self._config_input["tags"] = task_config_override.tags

            # if task_config_override.parents is not None:
            #     self._config_input["parents"] = task_config_override.parents

        # Returns an empty string to avoid productin incorrect sql
        return ""

    def setup(self):
        if self.needs_recompile:
            self.test_query = self.compiler.compile(self.task_config.file_name)
            self.test_query += " LIMIT 5\n"
        return Ok()

    def test(self):
        step_queries = {
            "Write Test Query": self.test_query,
            "Execute Test Query": self.test_query,
        }
        if self.run_arguments["command"] == "test":
            self.set_run_steps(list(step_queries.keys()))

        for step, query in step_queries.items():
            with self.step(step):
                if "Write" in step:
                    self.write_compilation_output(query, "test")
                if "Execute" in step:
                    try:
                        result = self.default_db.read_data(query)
                    except Exception as e:
                        return Exc(e)

        if len(result) == 0:
            return self.success()
        else:
            if self.run_arguments["debug"]:

                errinfo = f"Test failed. You can find the compiled test query at compile/{self.group}/{self.name}_test.sql"

                return self.fail(errinfo)
            else:
                return self.fail("Failed test types: custom")



from pathlib import Path

from jinja2 import Environment, FileSystemLoader, StrictUndefined


from ..utils.python_loader import PythonLoader
from ..logging.task_event_tracker import TaskEventTracker
from .task_wrapper import TaskWrapper
from .dummy import DummyTask
from .sql import SqlTask

from .autosql import AutoSqlTask
from .copy import CopyTask

# _excluded_properties = (
#     "name",
#     "type",
#     "tags",
#     "group",
#     "parents",
#     "sources",
#     "outputs",
#     "parameters",
#     "class",
#     "preset",
#     "on_fail",
# )

# def get_task_class(self, python_loader, task_type, config):
#     if task_type == "python":
#         return self.python_loader.get_class("python_tasks", config.get("class"))
#     elif task_type in _creators:
#         return Ok(_creators[task_type])
#     else:
#         return Err(
#             "task_type",
#             "invalid_task_type_error",
#             type=task_type,
#             group=config["group"],
#         )


class TaskBuilder:
    _creators = {}

    def __init__(self, python_folder, sql_folder, app_tracker, default_db, connections):
        self.app_tracker = app_tracker
        self.sql_folder = sql_folder
        self.default_db = default_db
        self.connections = connections

        if Path(python_folder).is_dir():
            self.python_loader = PythonLoader()
        else:
            self.python_loader = None

        self.base_jinja_env = Environment(
            loader=FileSystemLoader(str(Path("."))),
            undefined=StrictUndefined,
            keep_trailing_newline=True,
        )

    def get_task(self, config):
        if config.type == "dummy":
            runner_class = DummyTask
        elif config.type == "sql":
            runner_class = SqlTask
        elif config.type == "autosql":
            runner_class = AutoSqlTask
        elif config.type == "copy":
            runner_class = CopyTask
        elif config.type == "python":
            if self.python_loader is None:
                raise ValueError("No python module defined in the project")

        # (config.name, config.group, self.app_tracker.get_task_tracker(config.name), self.run_arguments, config.parameters, self.project_parameters, self.default_db, self.connections)

        return TaskWrapper(runner_class)


class SAYNError(Exception):
    pass



import csv
from pathlib import Path
import tempfile

from sqlalchemy import create_engine

from . import Database

db_parameters = [
    "account",
    "region",
    "user",
    "password",
    "database",
    "warehouse",
    "role",
    "schema",
]


class Snowflake(Database):
    def feature(self, feature):
        return feature in ("TABLE RENAME CHANGES SCHEMA",)

    def create_engine(self, settings):
        from snowflake.sqlalchemy import URL

        url_params = dict()
        for param in db_parameters:
            if param in settings:
                url_params[param] = settings.pop(param)

        return create_engine(URL(**url_params), **settings)

    def execute(self, script):
        conn = self.engine.connect()
        conn.connection.execute_string(script)
        conn.connection.commit()
        conn.connection.close()

    def _list_databases(self):
        """List the accessible databases for this connection."""
        databases = self.read_data("SHOW DATABASES;")
        databases = [db["name"].lower() for db in databases] + [""]
        return databases

    def _get_table_type(self, type):
        if type == "BASE TABLE":
            return "table"
        elif type == "VIEW":
            return "view"

    def _load_data_batch(self, table, data, schema, db):
        """Implements the load of a single data batch for `load_data`.

        Defaults to an insert many statement, but it's overloaded for specific
        database connector for more efficient methods.

        Args:
            table (str): The name of the target table
            data (list): A list of dictionaries to load
            schema (str): An optional schema to reference the table
        """
        full_table_name = f"{'' if db is None else db + '.'}{'' if schema is None else schema + '.'}{table}"
        template = self._jinja_env.get_template("snowflake_load_batch.sql")
        fname = "batch.csv"

        with tempfile.TemporaryDirectory() as tmpdirname:
            with (Path(tmpdirname) / fname).open("w") as f:
                writer = csv.DictWriter(
                    f, fieldnames=data[0].keys(), delimiter="\t", escapechar="\\"
                )
                writer.writeheader()
                writer.writerows(data)

            self.execute(
                template.render(
                    full_table_name=full_table_name,
                    temp_file_directory=tmpdirname,
                    temp_file_name=fname,
                )
            )



from __future__ import annotations

import re
from collections import deque
from typing import Mapping, Optional, Set

from . import Database


class DbObject:
    def __init__(
        self,
        compiler: DbObjectCompiler,
        connection_name: str,
        database: Optional[str],
        schema: Optional[str],
        table: Optional[str],
    ) -> None:
        self.compiler = compiler
        self.connection_name = connection_name
        self.database = database
        self.schema = schema
        self.table = table

        # The raw value is the value we use in our project code
        self.raw = ""
        if self.database is not None:
            self.raw += self.database + "."

        if self.schema is not None:
            self.raw += self.schema + "."

        if self.database is not None and self.schema is None:
            self.raw += "."

        if self.table is not None:
            self.raw += self.table

        # The key is prefixed with the connection name to make objects
        # unique across all databases
        self.key = f"{connection_name}:{self.raw}"

    def __hash__(self) -> int:
        return hash(self.key)

    def __eq__(self, obj: DbObject) -> bool:
        return self.key == obj.key

    def __lt__(self, obj: DbObject) -> bool:
        return self.key > obj.key

    def __repr__(self) -> str:
        return f"DbObject: {self.key}"


class DbObjectCompiler:

    regex_obj = re.compile(
        r"^\s*((?P<connection>[^:]+):)?((?P<c1>[^.]+)\.)?((?P<c2>[^.]+)\.)?(?P<c3>[^.]+)(?P<dots>\.{0,2})\s*$"
    )

    reference_level = {None: 0, "schema": -1, "db": -2}

    def __init__(
        self,
        connections: Mapping[str, Database],
        default_db: str,
        stringify: Mapping[str, Optional[str]],
        prod_stringify: Mapping[str, Optional[str]],
        from_prod: Set[str],
    ) -> None:
        self.connections = connections
        self.default_db = default_db
        self.from_prod = set()
        self.sources_from_prod = set()

        stringify = {k: v or prod_stringify[k] for k, v in stringify.items()}
        self.stringify = {
            t: self._get_stringify(
                t,
                stringify[f"{t}_prefix"],
                stringify[f"{t}_suffix"],
                stringify[f"{t}_override"],
            )
            for t in ("database", "schema", "table")
        }
        self.prod_stringify = {
            t: self._get_stringify(
                t,
                prod_stringify[f"{t}_prefix"],
                prod_stringify[f"{t}_suffix"],
                prod_stringify[f"{t}_override"],
            )
            for t in ("database", "schema", "table")
        }

        self.from_prod = {re.compile(o) for o in from_prod}

    def _get_stringify(
        self,
        type_: str,
        prefix: Optional[str],
        suffix: Optional[str],
        override: Optional[str],
    ) -> str:
        """Builds the final stringify"""

        # The default is a jinja template with the object name, meaning no modifications to the input
        stringify = "{" + type_ + "}"

        if override is not None:
            # if *_override is defined, use that
            stringify = override
        else:
            if prefix is not None and len(prefix) > 0:
                stringify = prefix + "_" + stringify

            if suffix is not None and len(suffix) > 0:
                stringify = stringify + "_" + suffix
        return stringify

    def set_sources_from_prod(self, sources_from_prod: Set[DbObject]) -> None:
        self.sources_from_prod = sources_from_prod

    def is_from_prod(self, obj: DbObject) -> bool:
        if obj.connection_name == self.default_db:
            # Only for src, we might want to produce the prod version of the object
            for regex in self.from_prod:
                # Only usage of the raw value of a db object.
                # We use raw as all specifications in the project code are meant
                # to follow sayn logic, rather than the db specific special nomenclature
                m = regex.match(obj.raw)
                if m is not None:
                    return True

            for source in self.sources_from_prod:
                if source == obj:
                    return True

        return False

    def _common_value(self, obj: DbObject, run_sensitive: bool) -> str:
        database = obj.database
        schema = obj.schema
        table = obj.table

        if obj.connection_name == self.default_db:
            if run_sensitive:
                is_prod = self.is_from_prod(obj)
            else:
                is_prod = False

            if database is not None:
                if is_prod:
                    database = self.prod_stringify["database"].format(database=database)
                else:
                    database = self.stringify["database"].format(database=database)

            if schema is not None:
                if is_prod:
                    schema = self.prod_stringify["schema"].format(schema=schema)
                else:
                    schema = self.stringify["schema"].format(schema=schema)

            if table is not None:
                if is_prod:
                    table = self.prod_stringify["table"].format(table=table)
                else:
                    table = self.stringify["table"].format(table=table)

        return self.connections[obj.connection_name]._obj_str(database, schema, table)

    def src_obj(self, obj: DbObject) -> DbObject:
        """Returns the transform object according to from_prod rules"""
        return self.from_string(
            self._common_value(obj, True), connection=obj.connection_name
        )

    def out_obj(self, obj: DbObject) -> DbObject:
        """Returns the transform object according to from_prod rules"""
        return self.from_string(
            self._common_value(obj, False), connection=obj.connection_name
        )

    def src_value(self, obj: DbObject) -> str:
        return self._common_value(obj, True)

    def out_value(self, obj: DbObject) -> str:
        return self._common_value(obj, False)

    def from_string(
        self, obj: str, connection: str | Database = None, level: str = None
    ) -> DbObject:
        """This is the entry point to db objects from a SAYN project. Accepting
        2 or 3 components is specified here.
        """
        match = self.regex_obj.match(obj)

        if match is None:
            raise ValueError(f'Incorrect format for database object "{obj}"')

        if level not in self.reference_level.keys():
            raise ValueError(f'Incorrect format for reference level "{level}"')

        if isinstance(connection, Database):
            in_connection_name = connection.name
        else:
            in_connection_name = connection

        groups = match.groupdict()

        if groups["connection"] is None:
            if connection is None:
                connection_name = self.default_db
            else:
                connection_name = in_connection_name
        else:
            if connection is None:
                connection_name = groups["connection"]
            else:
                if groups["connection"] != in_connection_name:
                    raise ValueError("Connection name mismatch")
                else:
                    connection_name = in_connection_name

        if groups["c1"] is not None and groups["c2"] is None:
            groups["c2"] = groups["c1"]
            groups["c1"] = None

        if self.reference_level[level] == 0:
            period_count = groups["dots"].count(".")

            if period_count > 2:
                raise ValueError(
                    f'Invalid number of periods (".") in database object {obj}'
                )

            provided_level = -1 * period_count
        else:
            provided_level = self.reference_level[level]

        component_elements = deque(
            [v for k, v in groups.items() if k != "connection" and k != "dots"]
        )

        component_elements.rotate(provided_level)

        components = dict(
            {"table": None, "schema": None, "database": None},
            **dict(
                zip(
                    ("table", "schema", "database"),
                    reversed(component_elements),
                )
            ),
        )

        if components["table"] is None and provided_level == 0:
            # This can't happen given the regexp, but this case avoids
            # static analysis errors
            raise ValueError("Error interpreting object string")

        return DbObject(
            self,
            connection_name,
            components["database"],
            components["schema"],
            components["table"],
        )



import inspect
from typing import Optional, List, Union
from enum import Enum


from ..database.unknown import UnknownDb
from .task import Task


class OnFailValue(str, Enum):
    skip = "skip"
    no_skip = "no_skip"


class PythonTask(Task):
    def config(self, **task_config):
        return self.success()

    def setup(self):
        return self.success()

    def run(self):
        return self.success()

    def compile(self):
        return self.success()

    def test(self):
        return self.success()


class DecoratorTask(PythonTask):
    def __init__(
        self,
        name,
        group,
        tracker,
        run_arguments,
        task_parameters,
        project_parameters,
        default_db,
        connections,
        compiler,
        src,
        out,
        sources,
        outputs,
        parents,
        tags,
        on_fail,
        func,
    ):
        super().__init__(
            name,
            group,
            tracker,
            run_arguments,
            task_parameters,
            project_parameters,
            default_db,
            connections,
            compiler,
            src,
            out,
        )

        self.compiler.update_globals(src=self.src, out=self.out)

        self._config_input["sources"].update(sources)
        self._config_input["outputs"].update(outputs)
        self._config_input["parents"].update(parents)
        self._config_input["tags"].update(tags)
        self._config_input["on_fail"] = on_fail
        self._func = func

    def config(self):
        self._has_tests = False

        while len(self._config_input["sources"]) > 0:
            self.src(self._config_input["sources"].pop())

        while len(self._config_input["outputs"]) > 0:
            self.out(self._config_input["outputs"].pop())

    def setup(self):
        pass

    def compile(self):
        pass

    def run(self):
        # Get the names of the arguments to the function
        sig = inspect.signature(self._func)

        wrapper_params = []
        for param in sig.parameters:
            if param == "context":
                # Special parameter equivalent to self
                wrapper_params.append(self)
            elif param in self.connections:
                if isinstance(self.connections[param], UnknownDb):
                    return self.fail(f'Connection "{param}" missing from settings')
                # The name of a connection makes it so that that argument
                # is linked to the connection object itself
                wrapper_params.append(self.connections[param])
            else:
                if param not in self.task_parameters:
                    value = None
                else:
                    value = self.task_parameters[param]
                # The rest are interpreted as task parameters
                wrapper_params.append(value)

        return self._func(*wrapper_params)

    def test(self):
        pass


class DecoratorTaskWrapper(Task):
    def __init__(
        self,
        func,
        sources: Optional[Union[List[str], str]] = None,
        outputs: Optional[Union[List[str], str]] = None,
        parents: Optional[Union[List[str], str]] = None,
        tags: Optional[Union[List[str], str]] = None,
        on_fail: Optional[OnFailValue] = None,
    ):
        """The init method collects the information provided by the decorator itself"""

        self.func = func

        # Get the names of the arguments to the function
        sig = inspect.signature(func)
        self.func_arguments = [p for p in sig.parameters]

        # Need to store these temporarily
        if sources is None:
            self.sources = set()
        elif isinstance(sources, str):
            self.sources = set()
            self.sources.add(sources)
        else:
            self.sources = sources

        if outputs is None:
            self.outputs = set()
        elif isinstance(outputs, str):
            self.outputs = set()
            self.outputs.add(outputs)
        else:
            self.outputs = outputs

        if parents is None:
            self.parents = set()
        elif isinstance(parents, str):
            self.parents = set()
            self.parents.add(parents)
        else:
            self.parents = parents

        if tags is None:
            self.tags = set()
        elif isinstance(tags, str):
            self.tags = set()
            self.tags.add(tags)
        else:
            self.tags = set(tags)

        if on_fail is None:
            self.on_fail = OnFailValue.skip
        elif on_fail in ("skip", "no_skip"):
            self.on_fail = OnFailValue[f"{on_fail}"]
        else:
            raise ValueError(f"{on_fail} not a valid value for on_fail field")

    def __call__(
        self,
        name,
        group,
        tracker,
        run_arguments,
        task_parameters,
        project_parameters,
        default_db,
        connections,
        compiler,
        src,
        out,
    ):
        task = DecoratorTask(
            name,
            group,
            tracker,
            run_arguments,
            task_parameters,
            project_parameters,
            default_db,
            connections,
            compiler,
            src,
            out,
            self.sources,
            self.outputs,
            self.parents,
            self.tags,
            self.on_fail,
            self.func,
        )

        return task


def task_type(func=None, sources=None, outputs=None, parents=None):
    if func:
        return DecoratorTaskWrapper(func)
    else:

        def wrapper(func):
            return DecoratorTaskWrapper(
                func,
                sources=sources,
                outputs=outputs,
                parents=parents,
            )

        return wrapper


def task(func=None, sources=None, outputs=None, parents=None, tags=None, on_fail=None):
    if func:
        return DecoratorTaskWrapper(func)
    else:

        def wrapper(func):
            return DecoratorTaskWrapper(
                func,
                sources=sources,
                outputs=outputs,
                parents=parents,
                tags=tags,
                on_fail=on_fail,
            )

        return wrapper



from halo import Halo

from .logger import Logger
from .log_formatter import LogFormatter, human


class FancyLogger(Logger):
    fmt = LogFormatter(use_colour=False, use_icons=False, output_ts=True)
    cfmt = LogFormatter(use_colour=True, use_icons=False, output_ts=True)
    spinner = Halo(spinner="dots")

    stage = None
    task = None
    step = None
    task_order = None
    total_tasks = None
    step = None
    step_order = None
    total_steps = None
    task_text = None
    step_text = None
    task_persist_msgs = list()

    def message(self, level, message, details):
        fmsg = self.cfmt.message(level, message, details)
        self.task_persist_msgs.append(fmsg)
        txt = f"{self.task_text}: "
        if self.step_text is not None:
            txt += f"{self.step_text}: "
        txt += ". ".join(fmsg["message"])
        self.spinner.text = txt

    def task_stage_finish(self, stage, duration, result):
        if stage != "config":
            if self.total_tasks == 0:
                self.spinner.text = f"{self.task} ({human(duration)})"
            else:
                self.spinner.text = f"[{self.task_order}/{self.total_tasks}] {self.task} ({human(duration)})"
        else:
            self.spinner.text = f" {self.task} ({human(duration)})"

        if result.is_ok:
            if stage in ("config", "setup"):
                self.spinner.clear()
                self.spinner.text_color = None
            else:
                self.spinner.text_color = "green"
                self.spinner.succeed()
                self.spinner.text_color = None

        else:
            if result.error.code == "parent_errors":
                self.spinner.text_color = "yellow"
                self.spinner.warn()
            elif (
                result.error.code == "setup_error"
                and result.error.details["status"].value == "skipped"
            ):
                self.spinner.text_color = "yellow"
                self.spinner.warn()
            else:
                self.spinner.text_color = "red"
                self.spinner.fail()

            self.spinner.text_color = None

            self.current_indent += 1

            message = self.fmt.error_result(duration, result.error)

            if self.step is not None:
                step_progress = f"{self.step_order}/{self.total_steps}"

                if isinstance(message["message"], list):
                    message["message"].append(self.fmt.red(f"On step {self.step}:\n"))
                    self.current_indent += 1
                    self.print(message)
                    self.current_indent -= 1
                else:
                    message["message"] = (
                        self.fmt.red(f"On step {step_progress} {self.step}: ")
                        + message["message"]
                    )
                    self.print(message)

            else:
                self.print(message)

            self.current_indent -= 1

        self.task = None
        self.task_order = None
        self.task_text = None

        self.current_indent += 1
        for msg in self.task_persist_msgs:
            if msg["level"] in ("warning", "error"):
                self.print(msg)
        self.current_indent -= 1

        self.task_persist_msgs = list()

    def report_event(self, context, event, stage, **details):
        if event == "message":
            self.message(details["level"], details["message"], details)

        elif context == "app":
            if event == "start_app":
                self.app_start(details)
                print()

            elif event == "finish_app":
                self.app_finish(details)

            elif event == "start_stage":
                self.stage = stage
                self.app_stage_start(stage, details)

            elif event == "finish_stage":
                self.spinner.stop()
                print()
                self.stage = None
                self.app_stage_finish(stage, details)
                print()

            else:
                self.unhandled(event, context, stage, details)

        elif context == "task":
            self.task = details["task"]
            self.task_order = details["task_order"]
            self.total_tasks = details["total_tasks"]
            ts = human(details["ts"])

            if event == "set_run_steps":
                # Less verbosity for this logger
                pass  # self.task_set_steps(details)

            elif event == "start_stage":
                if self.total_tasks == 0:
                    self.task_text = f"{self.task} (started at {ts})"
                else:
                    self.task_text = f"[{self.task_order}/{self.total_tasks}] {self.task} (started at {ts})"
                self.spinner.text = self.task_text
                self.spinner.start()

            elif event == "finish_stage":
                self.task_stage_finish(stage, details["duration"], details["result"])

            elif event == "start_step":
                self.step = details["step"]
                self.step_order = details["step_order"]
                self.total_steps = details["total_steps"]
                self.step_text = (
                    f"Step [{self.step_order}/{self.total_steps}] {self.step}"
                )
                self.spinner.text = f"{self.task_text}: {self.step_text}"

            elif event == "finish_step":
                if details["result"].is_ok:
                    self.step = None
                    self.step_order = None

                self.step_text = None

            else:
                self.unhandled(event, context, stage, details)

        else:
            self.unhandled(event, context, stage, details)

    def print(self, s=None):
        if s is None:
            pass
        else:
            prefix = "  " * self.current_indent
            s = s["message"]
            if isinstance(s, str):
                s = [s]
            elif not isinstance(s, list):
                raise ValueError("error in logging print")

            print(f"{prefix}{s[0]}")
            for e in s[1:]:
                for l in e.split("\n"):
                    print(f"{prefix}  {l}")



from collections import Counter
import datetime
import decimal
from itertools import groupby
from pathlib import Path
from typing import List, Optional, Union

from jinja2 import Environment, FileSystemLoader, StrictUndefined
from pydantic import BaseModel, validator, Extra
from sqlalchemy import MetaData, Table
from sqlalchemy.sql import sqltypes
from sqlalchemy import text

from ..core.errors import DBError, Exc, Ok


class Hook(BaseModel):
    sql: str


class Columns(BaseModel):
    name: str
    type: Optional[str]
    dst_name: Optional[str]
    description: Optional[str]

    class Tests(BaseModel):
        name: Optional[str]
        allowed_values: Optional[List[Union[int, float, bool, str]]]
        execute: bool = True

        class Config:
            extra = Extra.forbid

    tests: List[Union[str, Tests]] = list()

    class Config:
        extra = Extra.forbid


class BaseDDL(BaseModel):
    def base_ddl(self):
        columns = list()
        for c in self.columns:
            tests = list()
            for t in c.tests:
                if isinstance(t, str):
                    tests.append({"type": t, "allowed_values": [], "execute": True})
                else:
                    tests.append(
                        {
                            "type": t.name if t.name is not None else "allowed_values",
                            "allowed_values": t.allowed_values
                            if t.allowed_values is not None
                            else [],
                            "execute": t.execute,
                        }
                    )
            columns.append(
                {
                    "name": c.name,
                    "description": c.description,
                    "dst_name": c.dst_name,
                    "type": c.type,
                    "tests": tests,
                }
            )

        return {
            "columns": columns,
            "properties": list(),
            "post_hook": [h.dict() for h in self.post_hook],
        }


class DDL(BaseDDL):
    columns: List[Union[str, Columns]] = list()
    post_hook: List[Hook] = list()

    class Config:
        extra = Extra.forbid

    @validator("columns", pre=True)
    def transform_str_cols(cls, v):
        if v is not None and isinstance(v, List):
            return [{"name": c} if isinstance(c, str) else c for c in v]
        else:
            return v

    @validator("columns")
    def columns_unique(cls, v):
        dupes = {k for k, v in Counter([e.name for e in v]).items() if v > 1}
        if len(dupes) > 0:
            raise ValueError(f"Duplicate columns: {','.join(dupes)}")
        else:
            return v

    def get_ddl(self):
        return self.base_ddl()


class Database:
    """
    Base class for databases in SAYN.

    Databases are implemented using sqlalchemy, and the `engine` attribute is available
    when working in python tasks without the need for calling create_engine.

    Attributes:
        engine (sqlalchemy.Engine): A sqlalchemy engine referencing the database.
        name (str): Name of the db as defined in `required_credentials` in `project.yaml`.
        name_in_yaml (str): Name of db under `credentials` in `settings.yaml`.
        db_type (str): Type of the database.
        metadata (sqlalchemy.MetaData): A metadata object associated with the engine.
    """

    DDL = DDL

    def __init__(
        self,
        name,
        name_in_settings,
        db_type,
        common_params,
        settings,
    ):
        self.name = name
        self.name_in_settings = name_in_settings
        self.db_type = db_type
        self.max_batch_rows = common_params.get("max_batch_rows", 50000)
        self._settings = settings
        self._requested_objects = dict()

        self._jinja_env = Environment(
            loader=FileSystemLoader(Path(__file__).parent / "templates"),
            undefined=StrictUndefined,
            keep_trailing_newline=True,
        )
        self._jinja_test = Environment(
            loader=FileSystemLoader(Path(__file__).parent.parent / "tasks/tests"),
            undefined=StrictUndefined,
            keep_trailing_newline=True,
        )

    def _obj_str(
        self, database: Optional[str], schema: Optional[str], table: str
    ) -> str:
        return (
            f"{database if database is not None else ''}"
            f"{'.' if database is not None else ''}"
            f"{schema if schema is not None else ''}"
            f"{'.' if database is not None or (schema is not None) else ''}"
            f"{table if table is not None else ''}"
        )

    def _fully_qualify(
        self, database: Optional[str], schema: Optional[str], table: str
    ) -> str:
        return self._obj_str(database, schema, table)

    def feature(self, feature):
        # Supported sql_features
        #   - CREATE IF NOT EXISTS
        #   - CREATE TABLE NO PARENTHESES
        #   - INSERT TABLE NO PARENTHESES
        #   - DROP CASCADE
        #   - NO SET SCHEMA
        #   - NO ALTER INDEXES

        # CAN REPLACE TABLE
        # CAN REPLACE TABLE
        # CAN REPLACE VIEW
        # NEEDS CASCADE
        # CANNOT SPECIFY DDL IN SELECT
        # CANNOT ALTER INDEXES
        # CANNOT SET SCHEMA
        return feature in ()

    def create_engine(self, settings):
        raise NotImplementedError()

    def _activate_connection(self):
        self.engine = self.create_engine(self._settings)
        if self.engine is not None:
            # We'll have a None engine when the connection is missing from the settings.
            # We create said object only to allow the config stage to run correctly.
            self.metadata = MetaData(self.engine)
            # Force a query to test the connection
            self.execute("select 1")

    def _list_databases(self):
        raise NotImplementedError()

    def _list_objects(self, databases):
        """List the accessible databases for this connection."""
        objects = list()
        queries = ""
        for database in databases:
            if database != "":
                queries += f"""SELECT table_catalog, table_schema, table_name, table_type FROM {database}.INFORMATION_SCHEMA.TABLES"""
            else:
                queries += "SELECT '' AS table_catalog, table_schema, table_name, table_type FROM INFORMATION_SCHEMA.TABLES"
            queries += "\nUNION ALL\n"

        queries = "\n".join(queries.split("\n")[:-2])

        objects = {
            db.lower(): {
                schema.lower(): {
                    table.lower(): {
                        "type": self._get_table_type(type["table_type"]) for type in ggg
                    }
                    for table, ggg in groupby(gg, lambda x: x["table_name"])
                }
                for schema, gg in groupby(g, lambda x: x["table_schema"])
            }
            for db, g in groupby(
                sorted(
                    self.read_data(queries),
                    key=lambda x: (
                        x["table_catalog"],
                        x["table_schema"],
                        x["table_name"],
                    ),
                ),
                key=lambda x: x["table_catalog"],
            )
        }

        return objects

    def _get_table_type(self, type):
        if type.lower() == "table":
            return "table"
        elif type.lower() == "base table":
            return "table"
        elif type.lower() == "view":
            return "view"

    def _construct_tests_template(self, columns, table, test_file_name, schema):
        query = """
                   SELECT val
                        , col
                        , cnt
                        , type
                     FROM (
                """
        template = self._jinja_test.get_template(test_file_name)
        count_tests = 0
        breakdown = []
        for col in columns:
            tests = col["tests"]
            for t in tests:
                breakdown.append(
                    {
                        "column": col["name"]
                        if not col["dst_name"]
                        else col["dst_name"],
                        "type": t["type"],
                        "execute": t["execute"],
                        "allowed_values": ", ".join(
                            f"{format_type(c)}" for c in t["allowed_values"]
                        ),
                    }
                )

                if t["execute"]:
                    count_tests += 1
                    query += template.render(
                        **{
                            "table": table,
                            "schema": schema,
                            "name": col["name"]
                            if not col["dst_name"]
                            else col["dst_name"],
                            "type": t["type"],
                            "allowed_values": ", ".join(
                                f"{format_type(c)}" for c in t["allowed_values"]
                            ),
                        },
                    )

        parts = query.splitlines()[:-2]
        query = ""
        for q in parts:
            query += q.strip() + "\n"
        query += ") AS t\n LIMIT 5;"

        return count_tests, query, breakdown

    def _construct_tests(self, columns, table, schema=None):
        count_tests, query, breakdown = self._construct_tests_template(
            columns, table, "standard_tests.sql", schema
        )
        if count_tests == 0:
            return Ok([None, breakdown])

        return Ok([query, breakdown])

    def _validate_ddl(self, columns, table_properties, post_hook):
        if len(columns) == 0 and len(table_properties) == 0 and len(post_hook) == 0:
            properties = self.DDL().get_ddl()
        else:
            try:
                if table_properties is None or len(table_properties) == 0:
                    properties = self.DDL(
                        columns=columns,
                        post_hook=post_hook,
                    ).get_ddl()
                else:
                    properties = self.DDL(
                        columns=columns,
                        properties=table_properties,
                        post_hook=post_hook,
                    ).get_ddl()
            except Exception as e:
                return Exc(e, db=self.name, type=self.db_type)

        return self._format_properties(properties)

    def _format_properties(self, properties):
        if properties["columns"]:
            columns = []
            for col in properties["columns"]:
                entry = {
                    "name": col["name"],
                    "type": col["type"],
                    "dst_name": col["dst_name"],
                    "unique": False,
                    "not_null": False,
                    "allowed_values": False,
                }
                if "tests" in col:
                    entry.update({"tests": col["tests"]})
                    for t in col["tests"]:
                        if t["type"] != "values" and col["type"]:
                            entry.update({t["type"]: True})
                columns.append(entry)

            properties["columns"] = columns

        return Ok(properties)

    def test_problematic_values_template(
        self, failed: list, table: str, schema: str, test_file_name: str
    ) -> str:
        template = self._jinja_test.get_template(test_file_name)
        query = ""
        for f in failed:
            query += template.render(
                **{
                    "table": table,
                    "schema": schema,
                    "name": f[2],
                    "type": f[1],
                    "allowed_values": f[3],
                },
            )
        return query

    def test_problematic_values(self, failed: list, table: str, schema: str) -> str:
        return self.test_problematic_values_template(
            failed, table, schema, "standard_test_output.sql"
        )

    def _refresh_metadata(self, only=None, schema=None):
        """Refreshes the sqlalchemy metadata object.

        Args:
            only (list): A list of object names to filter the refresh on
            schema (str): The schema name to filter on the refresh
        """
        self.metadata.reflect(only=only, schema=schema, extend_existing=True)

    def _introspect(self, objects_to_introspect):
        out = dict()
        introspect_databases = set(objects_to_introspect.keys())
        databases = set(self._list_databases())
        schemata_list = [
            (db, schem)
            for db, schs in objects_to_introspect.items()
            for schem in list(schs.keys())
        ]
        introspect_schemata = set(schemata_list)
        databases_intersection = introspect_databases.intersection(databases)
        databases_intersection.add("")  # Always introspect the default db

        if len(databases_intersection) > 0:
            dbobjects = self._list_objects(databases_intersection)
            schemata = [
                (database, schema)
                for database, schemas in dbobjects.items()
                for schema in list(schemas.keys())
            ]
            schemata = set(schemata)
            schemata_intersection = introspect_schemata.intersection(schemata)

            if len(schemata_intersection) > 0:
                for db, schema in schemata_intersection:
                    if db not in out:
                        out[db] = dict()

                    if schema not in out[db]:
                        out[db][schema] = dict()

                    if db not in dbobjects or schema not in dbobjects[db]:
                        continue

                    tables = dbobjects[db][schema]

                    for table_name in objects_to_introspect[db][schema]:
                        if table_name in tables:
                            out[db][schema][table_name] = tables[table_name]
                        else:
                            out[db][schema][table_name] = {"type": None}

        self._requested_objects = out

    def _py2sqa(self, from_type):
        python_types = {
            int: sqltypes.BigInteger,
            str: sqltypes.Unicode,
            float: sqltypes.Float,
            decimal.Decimal: sqltypes.Numeric,
            datetime.datetime: sqltypes.TIMESTAMP,
            bytes: sqltypes.LargeBinary,
            bool: sqltypes.Boolean,
            datetime.date: sqltypes.Date,
            datetime.time: sqltypes.Time,
            datetime.timedelta: sqltypes.Interval,
            list: sqltypes.ARRAY,
            dict: sqltypes.JSON,
        }

        if from_type not in python_types:
            raise ValueError(f'Type not supported "{from_type}"')
        else:
            return python_types[from_type]().compile(dialect=self.engine.dialect)

    # API

    def execute(self, script):
        """Executes a script in the database. Multiple statements are supported.

        Args:
            script (sql): The SQL script to execute
        """
        with self.engine.connect().execution_options(autocommit=True) as connection:
            connection.execute(text(script))

    def read_data(self, query, **params):
        """Executes the query and returns a list of dictionaries with the data.

        Args:
            query (str): The SELECT query to execute
            params (dict): sqlalchemy parameters to use when building the final query as per
                [sqlalchemy.engine.Connection.execute](https://docs.sqlalchemy.org/en/13/core/connections.html#sqlalchemy.engine.Connection.execute)

        Returns:
            list: A list of dictionaries with the results of the query

        """
        if params is not None:
            res = self.engine.execute(query, **params)
        else:
            res = self.engine.execute(query)

        return [dict(zip([str(k) for k in res.keys()], r)) for r in res.fetchall()]

    def _read_data_stream(self, query, **params):
        """Executes the query and returns an iterator dictionaries with the data.

        The main difference with read_data() is that this method executes the query with a server-side
        cursor (sqlalchemy stream_results = True).

        Args:
            query (str): The SELECT query to execute
            params (dict): sqlalchemy parameters to use when building the final query as per
                [sqlalchemy.engine.Connection.execute](https://docs.sqlalchemy.org/en/13/core/connections.html#sqlalchemy.engine.Connection.execute)

        Returns:
            list: A list of dictionaries with the results of the query

        """
        with self.engine.connect().execution_options(stream_results=True) as connection:
            res = connection.execute(query, **params)
            fields = [str(k) for k in res.keys()]

            for record in res:
                yield dict(zip(fields, record))

    def _load_data_batch(self, table, data, schema, db):
        """Implements the load of a single data batch for `load_data`.

        Defaults to an insert many statement, but it's overloaded for specific
        database connector for more efficient methods.

        Args:
            table (str): The name of the target table
            data (list): A list of dictionaries to load
            schema (str): An optional schema to reference the table
        """
        table_def = self._get_table(table, schema)
        if table_def is None:
            raise DBError(
                self.name,
                self.db_type,
                f"Table {schema + '.' if schema is not None else ''}{table} does not exists",
            )

        with self.engine.connect().execution_options(autocommit=True) as connection:
            connection.execute(table_def.insert().values(data))

    def load_data(
        self, table, data, db=None, schema=None, batch_size=None, replace=False, **ddl
    ):
        """Loads a list of values into the database

        The default loading mechanism is an INSERT...VALUES, but database drivers
        will implement more appropriate methods.

        Args:
            table (str): The name of the target table
            data (list): A list of dictionaries to load
            schema (str): An optional schema to reference the table
            batch_size (int): The max size of each load batch. Defaults to
              `max_batch_rows` in the credentials configuration (settings.yaml)
            replace (bool): Indicates whether the target table is to be replaced
              (True) or new records are to be appended to the existing table (default)
            ddl (dict): An optional ddl specification in the same format as used
              in autosql and copy tasks

        Returns:
            int: Number of records loaded
        """
        batch_size = batch_size or self.max_batch_rows
        buffer = list()

        result = self._validate_ddl(
            ddl.get("columns", list()),
            ddl.get("table_properties", dict()),
            ddl.get("post_hook", list()),
        )

        if result.is_err:
            raise DBError(
                self.name,
                self.db_type,
                "Incorrect ddl provided",
                errors=result.error.details["errors"],
            )
        else:
            ddl = result.value

        check_create = True
        table_exists_prior_load = self._table_exists(table, schema)

        records_loaded = 0
        for i, record in enumerate(data):
            if check_create and (replace or not table_exists_prior_load):
                # Create the table if required
                if len(ddl.get("columns", list())) == 0:
                    # If no columns are specified in the ddl, figure that out
                    # based on the python types of the first record
                    columns = [
                        {"name": col, "type": self._py2sqa(type(val))}
                        for col, val in record.items()
                    ]
                    ddl = dict(ddl, columns=columns)

                query = self.create_table(
                    table, db=db, schema=schema, replace=replace, **ddl
                )
                self.execute(query)
                check_create = False

            if i % batch_size == 0 and len(buffer) > 0:
                self._load_data_batch(table, buffer, schema, db)
                records_loaded += len(buffer)
                buffer = list()

            buffer.append(record)

        if len(buffer) > 0:
            self._load_data_batch(table, buffer, schema, db)
            records_loaded += len(buffer)

        return records_loaded

    def _get_table(self, table, schema):
        """Create a SQLAlchemy Table object.

        Args:
            table (str): The table name
            schema (str): The schema or None

        Returns:
            sqlalchemy.Table: A table object from sqlalchemy
        """
        table_def = Table(table, self.metadata, schema=schema, extend_existing=True)

        if table_def.exists():
            table_def = Table(
                table, self.metadata, schema=schema, extend_existing=True, autoload=True
            )
            return table_def

    def _table_exists(self, table, schema):
        return self._get_table(table, schema) is not None

    # =========
    # ETL steps
    # =========

    # Intermediary steps

    def create_table(
        self,
        table,
        schema=None,
        db=None,
        select=None,
        replace=False,
        **ddl,
    ):
        db_name = db or ""
        schema_name = schema or ""
        full_name = fully_qualify(table, schema, db)
        if (
            db_name in self._requested_objects
            and schema_name in self._requested_objects[db_name]
            and table in self._requested_objects[db_name][schema_name]
        ):
            object_type = self._requested_objects[db_name][schema_name][table].get(
                "type"
            )
            table_exists = object_type == "table"
            view_exists = object_type == "view"
        else:
            table_exists = True
            view_exists = True

        template = self._jinja_env.get_template("create_table.sql")

        return template.render(
            table_name=table,
            full_name=full_name,
            view_exists=view_exists,
            table_exists=table_exists,
            select=select,
            replace=True,
            can_replace_table=self.feature("CAN REPLACE TABLE"),
            needs_cascade=self.feature("NEEDS CASCADE"),
            cannot_specify_ddl_select=self.feature("CANNOT SPECIFY DDL IN SELECT"),
            all_columns_have_type=len(
                [c for c in ddl.get("columns", dict()) if c.get("type") is not None]
            ),
            **ddl,
        )

    def merge_tables(
        self,
        src_table,
        dst_table,
        delete_key,
        cleanup=True,
        src_schema=None,
        dst_schema=None,
        src_db=None,
        dst_db=None,
        **ddl,
    ):
        src_table = fully_qualify(src_table, src_schema, src_db)
        dst_table = fully_qualify(dst_table, dst_schema, dst_db)

        template = self._jinja_env.get_template("merge_tables.sql")
        return template.render(
            dst_table=dst_table,
            src_table=src_table,
            cleanup=True,
            delete_key=delete_key,
        )

    def move_table(
        self,
        src_table,
        dst_table,
        src_schema=None,
        dst_schema=None,
        src_db=None,
        dst_db=None,
        **ddl,
    ):
        template = self._jinja_env.get_template("move_table.sql")
        dst_dbs = dst_db or ""
        dst_schemas = dst_schema or ""

        if (
            dst_dbs in self._requested_objects
            and dst_schemas in self._requested_objects[dst_dbs]
            and dst_table in self._requested_objects[dst_dbs][dst_schemas]
        ):
            object_type = self._requested_objects[dst_dbs][dst_schemas][dst_table].get(
                "type"
            )

            table_exists = bool(object_type == "table")
            view_exists = bool(object_type == "view")
        else:
            table_exists = True
            view_exists = True

        return template.render(
            table_exists=table_exists,
            view_exists=view_exists,
            src_db=src_db,
            src_schema=src_schema,
            src_table=src_table,
            dst_db=dst_db,
            dst_schema=dst_schema,
            dst_table=dst_table,
            cannot_alter_indexes=self.feature("CANNOT ALTER INDEXES"),
            needs_cascade=self.feature("NEEDS CASCADE"),
            rename_changes_schema=self.feature("TABLE RENAME CHANGES SCHEMA"),
            **ddl,
        )

    # ETL steps

    def replace_table(
        self,
        table,
        select,
        schema=None,
        db=None,
        tmp_schema=None,
        tmp_db=None,
        **ddl,
    ):
        # Create the temporary table
        can_replace_table = self.feature("CAN REPLACE TABLE")

        tmp_table = tmp_name(table)
        tmp_schema = tmp_schema or schema
        tmp_db = tmp_db or db

        if can_replace_table:
            create_or_replace = self.create_table(
                table, schema, db, select=select, replace=True, **ddl
            )

            return {"Create Or Replace Table": create_or_replace}

        else:
            create_or_replace = self.create_table(
                tmp_table, tmp_schema, tmp_db, select=select, replace=True, **ddl
            )

            # Move the table to its final location
            move = self.move_table(
                tmp_table,
                table,
                src_schema=tmp_schema or schema,
                dst_schema=schema,
                **ddl,
            )

            return {"Create Table": create_or_replace, "Move table": move}

    def replace_view(self, view, select, schema=None, db=None, **ddl):
        view_name = fully_qualify(view, schema, db)
        dbs = db or ""
        schemas = schema or ""
        if (
            dbs in self._requested_objects
            and schemas in self._requested_objects[dbs]
            and view in self._requested_objects[dbs][schemas]
        ):
            object_type = self._requested_objects[dbs][schemas][view].get("type")
            table_exists = object_type == "table"
            view_exists = object_type == "view"
        else:
            table_exists = True
            view_exists = True

        # ddl = self._format_properties(ddl).value

        template = self._jinja_env.get_template("create_view.sql")
        create = template.render(
            table_name=view_name,
            view_exists=view_exists,
            table_exists=table_exists,
            select=select,
            can_replace_view=self.feature("CAN REPLACE VIEW"),
            needs_cascade=self.feature("NEEDS CASCADE"),
            **ddl,
        )
        return {"Create View": create}

    def merge_query(
        self, table, select, delete_key, schema=None, db=None, tmp_schema=None, **ddl
    ):
        tmp_table = tmp_name(table)
        tmp_schema = tmp_schema or schema
        tmp_db = db

        create_or_replace = self.create_table(
            tmp_table, tmp_schema, tmp_db, select=select, replace=True
        )

        merge = self.merge_tables(
            tmp_table,
            table,
            delete_key,
            cleanup=True,
            src_schema=tmp_schema,
            dst_schema=schema,
            src_db=db,
            dst_db=db,
        )

        return {"Create Table": create_or_replace, "Merge Tables": merge}

    def get_db_object(self, database, schema, object):
        return self._object_builder.from_components(database, schema, object)


def fully_qualify(name, schema=None, db=None):
    return f"{db+'.' if db is not None else ''}{schema+'.' if schema is not None else ''}{name}"


def tmp_name(name):
    return f"sayn_tmp_{name}"


def format_type(value):
    if isinstance(value, str):
        return f"'{value}'"
    else:
        return value
